@@ -30,18 +30,27 @@ before "proxy-server" and add the following filter in the file:
     # Set control_exchange to publish to.
     control_exchange = swift
     # Set transport url
-    transport_url = rabbit://me:passwd@host:5672/virtual_host
+    url = rabbit://me:passwd@host:5672/virtual_host
     # set messaging driver
-    driver = messaging
+    driver = messagingv2
     # set topic
     topic = notifications
+    # skip metering of requests from listed project ids
+    ignore_projects = <proj_uuid>, <proj_uuid2>
+    # Whether to send events to messaging driver in a background thread
+    nonblocking_notify = False
+    # Queue size for sending notifications in background thread (0=unlimited).
+    # New notifications will be discarded if the queue is full.
+    send_queue_size = 1000
+    # Logging level control
+    log_level = WARNING
 """
 import functools
 import logging
 
 from oslo_config import cfg
-from oslo_context import context
 import oslo_messaging
+from oslo_utils import strutils
 from oslo_utils import timeutils
 from pycadf import event as cadf_event
 from pycadf.helper import api
@@ -49,7 +58,9 @@ from pycadf import measurement as cadf_measurement
 from pycadf import metric as cadf_metric
 from pycadf import resource as cadf_resource
 import six
+import six.moves.queue as queue
 import six.moves.urllib.parse as urlparse
+import threading
 
 _LOG = logging.getLogger(__name__)
 
@@ -98,15 +109,21 @@ class InputProxy(object):
 class Swift(object):
     """Swift middleware used for counting requests."""
 
+    event_queue = None
+    threadLock = threading.Lock()
+
     def __init__(self, app, conf):
         self._app = app
+        self.ignore_projects = [
+            proj.strip() for proj in
+            conf.get('ignore_projects', 'gnocchi').split(',')]
 
         oslo_messaging.set_transport_defaults(conf.get('control_exchange',
                                                        'swift'))
         self._notifier = oslo_messaging.Notifier(
             oslo_messaging.get_transport(cfg.CONF, url=conf.get('url')),
             publisher_id='ceilometermiddleware',
-            driver=conf.get('driver', 'messaging'),
+            driver=conf.get('driver', 'messagingv2'),
             topic=conf.get('topic', 'notifications'))
 
         self.metadata_headers = [h.strip().replace('-', '_').lower()
@@ -118,6 +135,28 @@ class Swift(object):
         if self.reseller_prefix and self.reseller_prefix[-1] != '_':
             self.reseller_prefix += '_'
 
+        _LOG.setLevel(getattr(logging, conf.get('log_level', 'WARNING')))
+
+        # NOTE: If the background thread's send queue fills up, the event will
+        #  be discarded
+        #
+        # For backward compatibility we default to False and therefore wait for
+        #  sending to complete. This causes swift proxy to hang if the
+        #  destination is unavailable.
+        self.nonblocking_notify = strutils.bool_from_string(
+            conf.get('nonblocking_notify', False))
+
+        # Initialize the sending queue and thread, but only once
+        if self.nonblocking_notify and Swift.event_queue is None:
+            Swift.threadLock.acquire()
+            if Swift.event_queue is None:
+                send_queue_size = int(conf.get('send_queue_size', 1000))
+                Swift.event_queue = queue.Queue(send_queue_size)
+                Swift.event_sender = SendEventThread(self._notifier)
+                Swift.event_sender.start()
+                _LOG.debug('Started sender thread')
+            Swift.threadLock.release()
+
     def __call__(self, env, start_response):
         start_response_args = [None]
         input_proxy = InputProxy(env['wsgi.input'])
@@ -156,6 +195,12 @@ class Swift(object):
 
     @_log_and_ignore_error
     def emit_event(self, env, bytes_received, bytes_sent, outcome='success'):
+        if ((env.get('HTTP_X_SERVICE_PROJECT_ID') or
+                env.get('HTTP_X_PROJECT_ID') or
+                env.get('HTTP_X_TENANT_ID')) in self.ignore_projects or
+                env.get('swift.source') is not None):
+            return
+
         path = urlparse.quote(env['PATH_INFO'])
         method = env['REQUEST_METHOD']
         headers = {}
@@ -169,8 +214,15 @@ class Swift(object):
 
         try:
             container = obj = None
-            version, account, remainder = path.replace(
-                '/', '', 1).split('/', 2)
+            path = path.replace('/', '', 1)
+            version, account, remainder = path.split('/', 2)
+        except ValueError:
+            try:
+                version, account = path.split('/', 1)
+                remainder = None
+            except ValueError:
+                return
+        try:
             if not version or not account:
                 raise ValueError('Invalid path: %s' % path)
             if remainder:
@@ -198,7 +250,7 @@ class Swift(object):
         # build object store details
         target = cadf_resource.Resource(
             typeURI='service/storage/object',
-            id=account.partition(self.reseller_prefix)[2])
+            id=account.partition(self.reseller_prefix)[2] or path)
         target.metadata = resource_metadata
         target.action = method.lower()
 
@@ -206,7 +258,8 @@ class Swift(object):
         initiator = cadf_resource.Resource(
             typeURI='service/security/account/user',
             id=env.get('HTTP_X_USER_ID'))
-        initiator.project_id = env.get('HTTP_X_TENANT_ID')
+        initiator.project_id = (env.get('HTTP_X_PROJECT_ID') or
+                                env.get('HTTP_X_TENANT_ID'))
 
         # build notification body
         event = cadf_event.Event(eventTime=now, outcome=outcome,
@@ -226,8 +279,38 @@ class Swift(object):
                 metric=cadf_metric.Metric(
                     name='storage.objects.outgoing.bytes', unit='B')))
 
-        self._notifier.info(context.get_admin_context().to_dict(),
-                            'objectstore.http.request', event.as_dict())
+        if self.nonblocking_notify:
+            try:
+                Swift.event_queue.put(event, False)
+                _LOG.debug('Event %s added to send queue', event.id)
+            except queue.Full:
+                _LOG.warning('Send queue FULL: Event %s not added', event.id)
+        else:
+            Swift.send_notification(self._notifier, event)
+
+    @staticmethod
+    def send_notification(notifier, event):
+        notifier.info({}, 'objectstore.http.request', event.as_dict())
+
+
+class SendEventThread(threading.Thread):
+
+    def __init__(self, notifier):
+        super(SendEventThread, self).__init__()
+        self.notifier = notifier
+        self.daemon = True
+
+    def run(self):
+        """Send events without blocking swift proxy."""
+        while True:
+            try:
+                _LOG.debug('Wait for event from send queue')
+                event = Swift.event_queue.get()
+                _LOG.debug('Got event %s from queue - now send it', event.id)
+                Swift.send_notification(self.notifier, event)
+                _LOG.debug('Event %s sent.', event.id)
+            except Exception:
+                _LOG.exception("SendEventThread loop exception")
 
 
 def filter_factory(global_conf, **local_conf):
